/*	Greg Tourville's Original Physics Engine
 *	Copyright 2006 Greg Tourville
 *	August 9th, 2006
 *	----------------------------
 *	By Greg aka XiGency
**/

#define	NULL		0L
//#define SLOW_WALLS

#include <math.h>
#include "physics.h"

inline float fastsqrt(float c) 
{ 
float output; 
asm("sqrt.s %0,%1" : "=f" (output) : "f" (c)); 
return output; 
} 


// Constructor
Physics::Physics()
{	// Fill Pointer Arrays with NULL's
	int i, x, y;
	for ( x = 0; x < MAX_SECTIONS_X; x++ )
		for ( y = 0; y < MAX_SECTIONS_Y; y++ )
			Sections[x][y] = NULL;

	for ( i = 0; i < MAX_CIRCLES; i++ )
		Circles[i] = NULL;

	for ( i = 0; i < MAX_WALLS; i++ )
		Walls[i] = NULL;

//	for ( i = 0; i < MAX_STATCIRCLES; i++ )
//		StatCircles[i] = NULL;

	for ( i = 0; i < MAX_HINGES; i++ )
		Hinges[i] = NULL;

	for ( i = 0; i < MAX_PARTICLES; i++ )
		Particles[i] = NULL;	

	// Set options
	FRICTION = 0.0f;
	GRAVITY = 0.0f;
	GRAVITY_ANGLE = 0.0f;
	LOSS_OF_ENERGY = 0.0f;

	nid = 0;
}

// Deconstructor
Physics::~Physics()
{	// Delete all none NULL pointers
	int i, x, y;
	for ( x = 0; x < MAX_SECTIONS_X; x++ )
		for ( y = 0; y < MAX_SECTIONS_Y; y++ )
			if ( Sections[x][y] != NULL )
				delete Sections[x][y];

	for ( i = 0; i < MAX_CIRCLES; i++ )
		if ( Circles[i] != NULL )
			delete Circles[i];

	for ( i = 0; i < MAX_WALLS; i++ )
		if ( Walls[i] != NULL )
			delete Walls[i];

//	for ( i = 0; i < MAX_STATCIRCLES; i++ )
//		if ( StatCircles[i] != NULL )
//			delete StatCircles[i];

	for ( i = 0; i < MAX_HINGES; i++ )
		if ( Hinges[i] != NULL )
			delete Hinges[i];

	for ( i = 0; i < MAX_PARTICLES; i++ )
		if ( Particles[i] != NULL )
			delete Particles[i];
}

int Physics::Initialize()
{
	int x, y, i, z;
	for ( x = 0; x < SectionsX; x++ )
		for ( y = 0; y < SectionsY; y++ )	// Make the sections
		{
			Sections[x][y] = new Section;

			// Clear it
			clearSection( Sections[x][y] );

			// Add Walls
			for ( z = 0; z < MAX_WALLS; z++ )
				if ( Walls[z] != NULL )
					// If its within this partition
					if ( Walls[z]->x1 >= (x*SectionWidth) && Walls[z]->x1 < ((x+1)*SectionWidth)
						&& Walls[x]->y1 >= (y*SectionHeight) && Walls[z]->y1 < ((y+1)*SectionHeight))
					
						// Find an empty one
						for ( i = 0; i < SECTION_WALLS; i++ )
							if ( Sections[x][y]->Walls[i] == NULL )
							{	// Add it
								Sections[x][y]->Walls[i] = Walls[z];
								break;
							}

			// Add Hinges
			for ( z = 0; z < MAX_HINGES; z++ )
				if ( Hinges[z] != NULL )
					// If its within this partition
					if ( Hinges[z]->x >= (x*SectionWidth) && Hinges[z]->x < ((x+1)*SectionWidth)
						&& Hinges[x]->y >= (y*SectionHeight) && Hinges[z]->y < ((y+1)*SectionHeight))
					
						// Find an empty one
						for ( i = 0; i < SECTION_HINGES; i++ )
							if ( Sections[x][y]->Hinges[i] == NULL )
							{	// Add it
								Sections[x][y]->Hinges[i] = Hinges[z];
								break;
							}
			
			// Add StatCircles

		}

	return 1;
}

int Physics::AddCircle( ColCircle* source )
{
	int i;
	for ( i = 0; i < MAX_CIRCLES; i++ )
	{
		if ( Circles[i] == NULL ) break;
	}

	if ( i < MAX_CIRCLES )
	{
		Circles[i] = source;
		Circles[i]->id = nid;
		nid++;
		return i;
	} else {
		return -1;	// Full
	}
}

int Physics::RemoveCircle( int index, bool del )			// By index
{
	if ( Circles[index] != NULL )
	{
		if ( del ) delete Circles[index];

		Circles[index] = NULL;
		return 1;
	} else {
		return -1;
	}
}

int Physics::AddWall( ColWall* source )			// Pass a struct* and added to list, returns index
{
	int i;
	for ( i = 0; i < MAX_WALLS; i++ )
	{
		if ( Walls[i] == NULL ) break;
	}

	if ( i < MAX_WALLS )
	{
		Walls[i] = source;
		return i;
	} else {
		return -1;	// Full
	}
}

int Physics::RemoveWall( int index, bool del )			// By index
{
	if ( Walls[index] != NULL )
	{
		// Find section
		int x, y, i;
		x = Walls[index]->x1 / SectionWidth;
		y = Walls[index]->y1 / SectionHeight;
		
		for ( i =0; i < SECTION_WALLS; i++ )
			if ( Sections[x][y]->Walls[i] == Walls[index] )
			{
				Sections[x][y]->Walls[i] = NULL;
//				Sections[x][y]->wallSort = true;
				break;
			}

		if ( del ) delete Walls[index];

		Walls[index] = NULL;
		return 1;
	} else {
		return -1;
	}
}

int Physics::AddHinge( ColHinge* source )
{
	int i;
	for ( i = 0; i < MAX_HINGES; i++ )
	{
		if ( Hinges[i] == NULL ) break;
	}

	if ( i < MAX_HINGES )
	{
		Hinges[i] = source;
		return i;
	} else {
		return -1;	// Full
	}
}

int Physics::RemoveHinge( int index, bool del )
{
	if ( Hinges[index] != NULL )
	{
		// Find section
		int x, y, i;
		x = Hinges[index]->x / SectionWidth;
		y = Hinges[index]->y / SectionHeight;
		
		for ( i =0; i < SECTION_HINGES; i++ )
			if ( Sections[x][y]->Hinges[i] == Hinges[index] )
			{
				Sections[x][y]->Hinges[i] = NULL;
//				Sections[x][y]->hingeSort = true;
				break;
			}

		if ( del ) delete Hinges[index];

		Hinges[index] = NULL;
		return 1;
	} else {
		return -1;
	}
}

int Physics::AddParticle( ColParticle* source )
{
	int i;
	for ( i = 0; i < MAX_PARTICLES; i++ )
	{
		if ( Particles[i] == NULL ) break;
	}

	if ( i < MAX_PARTICLES )
	{
		Particles[i] = source;
		return i;
	} else {
		return -1;	// Full
	}
}

/*
int Physics::AddStatCircle( ColStatCircle* source )
{
	int i;
	for ( i = 0; i < MAX_STATCIRCLES; i++ )
	{
		if ( StatCircles[i] == NULL ) break;
	}

	if ( i < MAX_PARTICLES )
	{
		StatCircles[i] = source;
		return i;
	} else {
		return -1;	// Full
	}
}

int Physics::RemoveStatCircle( int index, bool del );
{
	if ( StatCircles[index] != NULL )
	{
		if ( del ) delete StatCircles[index];

		StatCircles[index] = NULL;
		return 1;
	} else {
		return -1;
	}
}
*/


/* ///////////	UPDATE	/////////////////	*/
int Physics::GlobalUpdate( double dt )
{
	CircleUpdate( dt );

	return 1;
};

	float	x, y;
	float	xsp, ysp;
	float	rot, rotsp;
	int		id;
	int		size;

int Physics::CircleUpdate( double dt )
{
	// CIRCLE UPDATE ///////////////////////
	// & BOUNDRIE COLLISION ////////////////
	int i, z, x, y;
	for ( i = 0; i < MAX_CIRCLES; i++ )
	{
		if ( Circles[i] != NULL )
		{
			circleUpdate( Circles[i], dt );
		}
	}

#ifndef SLOW_WALLS
	// WALL COLLISIONS ////////////////////////////////
	for ( i = 0; i < MAX_CIRCLES; i++ )
	{
		if ( Circles[i] != NULL )
		{
			int firstX, firstY;
			int lastX, lastY;

			firstX = (int)(Circles[i]->x/SectionWidth);
			lastX = firstX+1;
			if ( firstX > 0 ) firstX--;
			//if ( firstX > 0 ) firstX--;
			if ( lastX < (SectionsX) ) lastX++;
			//if ( lastX < (SectionsX) ) lastX++;

			firstY = (int)(Circles[i]->y/SectionHeight);
			lastY = firstY+1;
			if ( firstY > 0 ) firstY--;
			//if ( firstY > 0 ) firstY--;
			if ( lastY < (SectionsY) ) lastY++;
			//if ( lastY < (SectionsY) ) lastY++;

			for ( x = firstX; x < lastX; x++ )
				for ( y = firstY; y < lastY; y++ )
					for ( z = 0; z < SECTION_WALLS; z++ )
						if ( Sections[x][y]->Walls[z] != NULL )
						{
							circleWallCollision( Circles[i], Sections[x][y]->Walls[z] );
						}



		}
	}
#else
	// WALL COLLISIONS ////////////////////////////////
	for ( i = 0; i < MAX_CIRCLES; i++ )
	{
		if ( Circles[i] != NULL )
		{
			for ( z = 0; z < MAX_WALLS; z++ )
				if ( Walls[z] != NULL )
				{
					circleWallCollision( Circles[i], Walls[z] );
				}

		}
	}
#endif



	// Circle Collisions	///////////////////////////////
	for ( i = 0; i < MAX_CIRCLES; i++ )
	{
		if ( Circles[i] != NULL )
		{
			for ( z = 0; z < MAX_CIRCLES; z++ )
			{
				if ( Circles[z] != NULL && z != i )
				{
					ColCircle* r = Circles[z];
					ColCircle* b = Circles[i];

					circleCircleCollision( r, b );
				}
			}
		}
	}

	return 1;
};



// Auxiliary Functions
int	clearSection( Section* s )
{
	int i;
	for ( i = 0; i < SECTION_WALLS; i++ )
		s->Walls[i] = NULL;

	for ( i = 0; i < SECTION_HINGES; i++ )
		s->Hinges[i] = NULL;

//	for ( i = 0; i < SECTION_CIRCLES; i++ )
//		s->CIRCLES[i] = NULL;

	return 1;
}


int Physics::circleUpdate( ColCircle* b, double dt )
{
	// Update Forces
	b->x += (b->xsp*dt);
	b->y += (b->ysp*dt);

#ifdef CIRC_FAKE_ROT
	b->rotsp = -b->xsp;
#endif

	b->rot += (b->rotsp*dt);

	// Frictions
	if ( FRICTION != 0.0f )
	{
		b->xsp -= (b->xsp * FRICTION * dt);
		b->ysp -= (b->ysp * FRICTION * dt);
		b->rotsp -= (b->rotsp * FRICTION * dt);
	}

	// Gravity
	if ( GRAVITY != 0.0f )
	{
		b->xsp += sin( GRAVITY_ANGLE * toRads ) * GRAVITY;
		b->ysp += cos( GRAVITY_ANGLE * toRads ) * GRAVITY;
	}

	// Outer Boundries
	if ( b->x - b->size < LEFT_X )
	{
		b->x = (float)(LEFT_X  + b->size);
		if ( b->xsp < 0.0f )
			b->xsp = -b->xsp * (1.0f-LOSS_OF_ENERGY);
	} else if ( b->x  + b->size > RIGHT_X ) {
		b->x = (float)(RIGHT_X - b->size);
		if ( b->xsp > 0.0f )
			b->xsp = -b->xsp * (1.0f-LOSS_OF_ENERGY);
	}
	if ( b->y - b->size < UPPER_Y )
	{
		b->y = (float)(UPPER_Y + b->size);
		if ( b->ysp < 0.0f )
			b->ysp = -b->ysp * (1.0f-LOSS_OF_ENERGY);
	} else if ( b->y + b->size > LOWER_Y ) {
		b->y = (float)(LOWER_Y - b->size);
		if ( b->ysp > 0.0f )
			b->ysp = -b->ysp * (1.0f-LOSS_OF_ENERGY);
	}

	return 1;
}


int Physics::circleWallCollision( ColCircle* b, ColWall* w )
{
	// Possible Collision

	// Point 1

	int d = w->size + b->size;
	float p1Dist = fastsqrt( (float)(( w->x1 - b->x )*( w->x1 - b->x ) + ( w->y1 - b->y )*( w->y1 - b->y )) );

	// Definitive Check
	if ( p1Dist < d )
	{
		double angle = atan2( b->x - w->x1, b->y - w->y1 );
		
		b->x = w->x1 + sin(angle) * d;
		b->y = w->y1 + cos(angle) * d;
	}
	

	// Point 2

	// Quick Check
	if ( (fabs( w->x2 - b->x ) + fabs( w->y2 - b->y ) ) < (d*1.5f) )
	{	// Definitive Check
		if ( ( ( w->x2 - b->x )*( w->x2 - b->x ) + ( w->y2 - b->y )*( w->y2 - b->y ) ) < (d*d) )
		{
			double angle = atan2( b->x - w->x2, b->y - w->y2 );
			
			b->x = w->x2 + sin(angle) * d;
			b->y = w->y2 + cos(angle) * d;
		}
	}

	// Line
	double ang1, ang2, angL1, angL2, dif1, dif2, ang, shortDist;
	ang1 = (atan2(b->x-w->x1,b->y-w->y1));
	ang2 = (atan2(b->x-w->x2,b->y-w->y2));
	angL1 = (atan2((float)(w->x2-w->x1),(float)(w->y2-w->y1)));
	angL2 = (atan2((float)(w->x1-w->x2),(float)(w->y1-w->y2)));
	dif1 = fabs(angL1 - ang1);
	dif2 = fabs(angL2 - ang2);

	// Is it between the two points
	if ( (cos(dif1) >= 0.0f) && (cos(dif2) >= 0.0f) )
	{
		ang = dif1;

		shortDist = fabs(sin(ang) * p1Dist);

		// Collision
		if ( shortDist < d )
		{
			float cx, cy;
			// Collision!
			float lineDist = cos(ang) * p1Dist;
			cx = w->x1 + (sin(angL1) * lineDist);
			cy = w->y1 + (cos(angL1) * lineDist);

			ang = atan2(b->x-cx,b->y-cy);

			// Fix the position
			b->x = cx + sin(ang)*(d);
			b->y = cy + cos(ang)*(d);

			// Reflect
			double Angle1a, Angle1b, speed;
			Angle1a = atan2(b->xsp, b->ysp);
			Angle1b = ( (angL1*2) - (Angle1a) );
			speed = fastsqrt( (float)((b->xsp*b->xsp) + (b->ysp*b->ysp)) );
			b->xsp = sin(Angle1b) * speed * (1.0f - LOSS_OF_ENERGY);
			b->ysp = cos(Angle1b) * speed * (1.0f - LOSS_OF_ENERGY);

			// Check for roll++


		}
	}

	return 1;
}


int Physics::circleCircleCollision( ColCircle* r, ColCircle* b )
{
	int d = b->size + r->size;

	// Quick Check
	if (( fabs( b->x - r->x ) + fabs( b->y - r->y ) ) < (d*1.5f))
	{
		// Definite check
		float distance = ( ( b->x - r->x )*( b->x - r->x ) + ( b->y - r->y )*( b->y - r->y ) );
		if ( distance < (d*d) )
		{
			// Fix overlap
			float centerX, centerY;
			centerX = (b->x + r->x)/2;
			centerY = (b->y + r->y)/2;

			float ang = atan2(b->x - centerX, b->y - centerY);
			float hd = (float)(d)/2;
			if ( b->kind + r->kind == 3 ) 
			{
				b->kind = 0;	// You've been hit
				hd = hd * 3;	// Make the guy and goose fly apart
			}			

			b->x = centerX + sin(ang)*hd;
			b->y = centerY + cos(ang)*hd;
			
			r->x = centerX - sin(ang)*hd;
			r->y = centerY - cos(ang)*hd;

			// Bounce\Reflect
/*
			float ang1, ang2, theAngle, Direct1, Direct2;

			// Directness 1
			ang1 = atan2(b->xsp, b->ysp);
			ang2 = atan2(r->x-b->x, r->y-b->y);
			theAngle = fabs(ang1 - ang2);
			Direct1 = cos(theAngle);
			if (Direct1 < 0.0f) Direct1 = 0;
			
			// Directness 2
			ang1 = atan2(r->xsp, r->ysp);
			ang2 = atan2(b->x-r->x, b->y-r->y);
			theAngle = fabs(ang1 - ang2);
			Direct2 = cos(theAngle);
			if (Direct2 < 0.0f) Direct1 = 0;

			// Reflection
			float Angle1a, Angle2a, AngleL, Angle1b, Angle2b;
			
			Angle1a = atan2(b->xsp, b->ysp);
			Angle2a = atan2(r->xsp, r->ysp);

			AngleL = atan2(b->y-r->y, b->x-r->x);

			Angle1b = (AngleL * 2 - Angle1a);
			Angle2b = (AngleL * 2 - Angle2a);

			float Energy1, Energy2, Energy1Old, Energy2Old, Energy1New, Energy2New;

			Energy1 = fastsqrt( (float)(b->xsp*b->xsp + b->ysp*b->ysp) ) * b->mass;
			Energy2 = fastsqrt( (float)(r->xsp*r->xsp + r->ysp*r->ysp) ) * r->mass;

			Energy1Old = Energy1 * (1-Direct1);
			Energy2Old = Energy2 * (1-Direct2);
			Energy1New = (Energy2 * Direct2 * LOSS_OF_ENERGY);
			Energy2New = (Energy1 * Direct1 * LOSS_OF_ENERGY);

			b->xsp = (sin(Angle1b) * Energy1New / b->mass) + (sin(Angle1a) * Energy1Old / b->mass);
			b->ysp = (cos(Angle1b) * Energy1New / b->mass) + (cos(Angle1a) * Energy1Old / b->mass);

			r->xsp = (sin(Angle2b) * Energy2New / r->mass) + (sin(Angle2a) * Energy2Old / r->mass);
			r->ysp = (cos(Angle2b) * Energy2New / r->mass) + (cos(Angle2a) * Energy2Old / r->mass);
*/

		}
	}

	return 1;
}
